'**********************************************************************************************
'*  io.spin - handle input, output and break requests sent by qz80
'*
'*  (c) Juergen Buchmueller <pullmoll@t-online.de>
'*
'* $Id: io.spin,v 1.3 2010-04-12 12:18:38 pm Exp $
'**********************************************************************************************
CON

' Define this to check for contiguous image files
#define	CHECK_CONTIG

' Define this to make 128 byte record writes pending for 1 second
#define TIMED_FLUSH

	io_cmd_out		=	$01		' I/O emulator command output
	io_cmd_in		=	$02		' I/O emulator command input
	io_cmd_break		=	$03		' I/O emulator command break
	io_cmd_kbd		=	$04		' I/O emulator command read (TRS-80 keyboard)

	constat			=	$10		' sio port 1 status port
	condata			=	$11		' sio port 1 data port
	punstat			=	$12		' sio port 2 status port
	pundata			=	$13		' sio port 2 data port

	myname_port		=	$70		' hopefully not used by anything else
	port2_Off		=	$71		' a little CP/M program sends an OUT to port &H71
	Forecolor		=	$72		' set foreground color
	Backcolor		=	$73		' set background color
	keystat			=	$74		' test just keyboard status (constat does this, too)
	keydata			=	$75		' get just the keyboard character

	port2_io		=	$78		' for input and output to port 2
	port2_test		=	$7D		' to test if a byte is ready
	port2_baud		=	$7C		' set baud rate multiple of 1200 eg 1=1200, 16=19200, 96=115200

	hdsk_reset		=	$1		' command to reset controller
	hdsk_read		=	$2		' read command
	hdsk_write		=	$3		' write command
	hdsk_param		=	$4		' get hard disk parameters
	hdsk_port		=	$fd		' control port for simulated hard disk

	simh_setbankselect	=	12		' SIMH command to set memory bank
	simh_reset		=	14		' SIMH reset command
	simh_hasbankedmemory	=	18		' SIMH command to check for banked memory
	simh_stoptimer		=	22		' SIMH stop timer interrupt
	simh_port		=	$fe		' control port for simulated hard disk

	baud			=	19200		' baud rate

	dir_none		=	0		' no i/o direction selected
	dir_out			=	1		' expecting client to write to the port
	dir_in			=	2		' expecting client to read from the port

#ifdef HAVE_TV
	SCREEN_W		=	80
	SCREEN_H		=	25
#endif HAVE_TV
#ifdef HAVE_VGA
	SCREEN_W		=	80
	SCREEN_H		=	40
#endif HAVE_VGA

VAR
	long	cog
	long	params[8]
	long	io_command
#ifdef HAVE_TV
	long	vt_command
#endif HAVE_TV
#ifdef HAVE_VGA
	long	vt_command
	long	vga_vsync
	word	vga_colors[64]			' max 64 rows
	byte	vga_cursor[8]
#endif HAVE_VGA
	byte	name[12+1]
	byte	datetime[20+1]

OBJ
	fat	: "fatfs"
	sio	: "pcFullDuplexSerial4FC"
#ifdef HAVE_KBD
	kbd	: "Keyboard"
#endif HAVE_KBD
#ifdef HAVE_TV
	tv80	: "tv80"
	vt100	: "vt100"
#endif
#ifdef HAVE_VGA
	vga	: "VGA_HiRes_Text"
	vt100	: "vt100"
#endif HAVE_VGA

PUB start(pins) : okay | i, dent, sec, size
	stop

	' This object has to be first, or it just won't work... talking about taking speed to the limit!?
	okay := \fat.start(iobuff_ptr, byte[pins][4], byte[pins][5], byte[pins][6], byte[pins][7])
	if okay < 0
		return -1
	pdo_command := fat.get_SPI_command
	pdo_blocknr := pdo_command + 4
	pdo_address := pdo_command + 8

	sio.AddPort(0, byte[pins][0], byte[pins][1], -1, -1, 0, 0, baud)
	if byte[pins][2] <> $ff
		sio.AddPort(1, byte[pins][2], byte[pins][3], -1, -1, 0, 0, baud)

	okay := sio.Start
	if okay =< 0
		return
	success(string("SIO"))
	tx_head0 := sio.get_tx_head
	tx_tail0 := sio.get_tx_tail
	tx_buffer0 := sio.get_tx_buffer
	rx_head0 := sio.get_rx_head
	rx_tail0 := sio.get_rx_tail
	rx_buffer0 := sio.get_rx_buffer

	if byte[pins][2] <> $ff
		tx_head1 := sio.get_tx_head1
		tx_tail1 := sio.get_tx_tail1
		tx_buffer1 := sio.get_tx_buffer1
		rx_head1 := sio.get_rx_head1
		rx_tail1 := sio.get_rx_tail1
		rx_buffer1 := sio.get_rx_buffer1

#ifdef HAVE_KBD
	if byte[pins][8] == $ff or byte[pins][9] == $ff
		return -1
	okay := \kbd.startx(byte[pins][8], byte[pins][9], %100, 40)
	if okay < 0
		failure(string("KBD"),okay)
		return -1
	success(string("KBD"))
	kbd_tail := kbd.tail_ptr
	kbd_head := kbd.head_ptr
	kbd_keys := kbd.keys_ptr
#endif

#ifdef HAVE_VGA
	vt_cmdptr := @vt_command
	if byte[pins][10] == $ff
		return -1
	okay := \vga.start(byte[pins][10], @vga_screen, @vga_colors, @vga_cursor, @vga_vsync)
	if okay <> TRUE
		return -1
	wordfill(@vga_colors, %%0020_3300, 64)
	vga_cursor[0] := 0		' cursor x
	vga_cursor[1] := 0		' cursor y
	vga_cursor[2] := %110		' on, blink slow
	vt_cmdptr := @vt_command
#endif HAVE_VGA

#ifdef HAVE_TV
	if byte[pins][11] == $ff
		return -1
	params[0] := byte[pins][11]
	params[1] := @tv_screen		' TV screen buffer
	okay := tv80.start(@params)
	if okay =< 0
		failure(string("TV 80"),okay)
		return
	success(string("TV80"))
	vt_cmdptr := @vt_command
#endif HAVE_TV

	str(string("qZ80 I/O starting...",13,10))

	okay := \fat.mount
	if okay < 0
		failure(string("FAT mount"),okay)
		return
	str(string("Volume serial #"))
	hex(\fat.get_serialno >> 16, 4)
	chr("-")
	hex(\fat.get_serialno, 4)
	str(string(", label "))
	str(\fat.get_volume_label)
	crlf

	read_boot(@boot_name)
	repeat i from 0 to 7
		map_disk(@lba_drive_a + i << 2, drive_name[i])

	okay := cog := COGNEW(@entry, @io_command) + 1
	if okay =< 0
		failure(string("I/O"),okay)
		return -1
	success(string("I/O"))


#ifdef HAVE_VGA
	params[0] := @vt_command
	params[1] := @vga_screen	' VGA screen buffer
	params[2] := @vga_colors	' VGA row colors
	params[3] := @vga_cursor	' VGA cursor positions
	params[4] := @vga_vsync		' VGA vertical sync flag
	params[5] := SCREEN_W		' terminal width
	params[6] := SCREEN_H		' terminal height
	okay := \vt100.start(@params)
	if okay =< 0
		failure(string("VT100 emulation"),okay)
		return
#endif

#ifdef HAVE_TV
	params[0] := @vt_command
	params[1] := @tv_screen		' TV screen buffer
	params[2] := 0			' no colors
	params[3] := 0			' no cursor
	params[4] := 0			' no vsync
	params[5] := SCREEN_W		' terminal width
	params[6] := SCREEN_H		' terminal height
	okay := \vt100.start(@params)
	if okay =< 0
		failure(string("VT100"),okay)
		return
	success(string("VT100"))
#endif
	return 0

PUB stop
'' Stop CPU simulator - frees a cog
	if cog
		cogstop(cog~ - 1)

PUB command_addr
'' Return the address of the 3 longs block
	return @io_command

PUB chr(ch)
	sio.tx(0,ch)

PUB str(str_ptr) | i
	repeat i from 0 to strsize(str_ptr)-1
		chr(byte[str_ptr][i])

PUB crlf
	chr(13)
	chr(10)

PUB dec_size(value)
	if value > 1048576
		dec(value/1048576)
		chr(".")
		dec(10*(value // 1048576) / 1048576)
		str(string("MB"))
	elseif value > 1024
		dec(value/1024)
		chr(".")
		dec(10*(value // 1024) / 1024)
		str(string("KB"))
	else
		dec(value)

PUB dec(value) | _i
'' Print a decimal number
	if value < 0
		-value
		chr("-")

	_i := 1_000_000_000

	repeat 10
		if value => _i
			chr(value / _i + "0")
			value //= _i
			result~~
		elseif result or _i == 1
			chr("0")
		_i /= 10

PUB hex(value, digits)
'' Print a hexadecimal number
	value <<= (8 - digits) << 2
	repeat digits
		chr(lookupz((value <-= 4) & $F : "0".."9", "A".."F"))

PUB bin(value, digits)
'' Print a binary number
	value <<= 32 - digits
	repeat digits
		chr((value <-= 1) & 1 + "0")

PRI failure(who, val)
	str(who)
	str(string(" failed: "))
	dec(val)
	crlf

PRI success(who)
	str(who)
	str(string(" initialized, "))
	dec(free_cogs)
	str(string(" cogs free.", 13, 10))

PUB free_cogs : free | i, cogs[8], jmp_0
	jmp_0 := %010111_0001_1111_000000000_000000000
	repeat while (cogs[free] := cognew(@jmp_0, 0)) => 0 
		free++
	if free
		repeat i from 0 to free - 1
			cogstop(cogs[i])
	return free

PRI read_boot(fname_ptr) | dent, size, sec
	dent := fat.find_file(fname_ptr)
	if dent == 0
		failure(fname_ptr,dent)
		return -1
	size := fat.get_dent_size(dent)
	sec := fat.get_dent_first_sector(dent)
	fat.get_dent_name(dent, @name)
	fat.get_dent_date_time(dent, @datetime)
	str(@name)
	str(string(", sector "))
	hex(sec,6)
	str(string(", size "))
	dec_size(size)
	str(string(", "))
	str(@datetime)
	crlf
	fat.read_sector(sec)
	fat.write_xmm_ram($FF00, 256)
	long[iobuff_ptr] := $00FF00C3
	fat.write_xmm_ram($0000, 4)


PRI map_disk(lbaptr, fname_ptr) : okay | dent, size, clus, exp, err, sec
	dent := fat.find_file(fname_ptr)
	if dent < 0
		failure(fname_ptr, dent)
		return -1
	sec := fat.get_dent_first_sector(dent)
	size := fat.get_dent_size(dent)
	long[lbaptr] := sec
	fat.get_dent_name(dent, @name)
	fat.get_dent_date_time(dent, @datetime)
	str(@name)
	str(string(", sector "))
	hex(sec, 6)
	str(string(", size "))
	dec_size(size)
	str(string(", "))
	str(@datetime)
#ifdef CHECK_CONTIG
	clus := fat.get_dent_cluster(dent)
	exp := clus + 1
	err := 0
	repeat while (clus := fat.fat_next(clus)) => 0
		if clus <> exp
			err++
			quit
		exp := exp + 1
	if err
		str(string(" not contiguous!"))
	else
		str(string(" contiguous - okay."))
#endif
	crlf

DAT
'---------------------------------------------------------------------------------------------------------
		org	0
'---------------------------------------------------------------------------------------------------------
#ifdef HAVE_VGA
vga_screen	long
#endif HAVE_VGA
#ifdef HAVE_TV
tv_screen	long
#endif HAVE_TV
'---------------------------------------------------------------------------------------------------------
entry
iocmd_ptr	mov	iocmd_ptr, PAR
iocmd		mov	iobuff_lba, minus1
		jmp	#startup

' pcFullDuplexSerial4FC
tx_head0	long	0-0
tx_tail0	long	0-0
tx_buffer0	long	0-0
rx_head0	long	0-0
rx_tail0	long	0-0
rx_buffer0	long	0-0

tx_head1	long	0-0
tx_tail1	long	0-0
tx_buffer1	long	0-0
rx_head1	long	0-0
rx_tail1	long	0-0
rx_buffer1	long	0-0

sio_head	long	0
sio_tail	long	0
sio_temp	long	0

#ifdef HAVE_KBD
kbd_tail	long	0-0
kbd_head	long	0-0
kbd_keys	long	0-0
kbd_ctrl	long	1<<9
#endif HAVE_KBD

#ifdef HAVE_TV
vt_cmdptr	long	0-0
#endif

#ifdef HAVE_VGA
vt_cmdptr	long	0-0
#endif

pdo_command	long	0-0
pdo_blocknr	long	0-0		' SD block number for r/w commands
pdo_address	long	0-0		' hub RAM address for r/w commands
iobuff_ptr	long	@@@iobuff

minus1		long	$ffffffff
low_24bits	long	$00ffffff
low_16bits	long	$0000ffff
h128		long	128<<16
msg_break_ptr	long	@@@msg_break

addr		long	0
port		long	0
data		long	0
null		long	0

simh_command	long	0		' currently active command
simh_status	long	0		' status byte
simh_params	long	@@@simh_params_	' pointer to the simh io buffer
simh_params_ptr	long	0		' pointer into simh_params
simh_params_cnt	long	0		' count of parameters expected

hdsk_command	long	0
hdsk_status	long	0
hdsk_lba_drives	long	@@@lba_drive_a
hdsk_drive	long	@@@hdsk_drive_	' parameters of read/write command
hdsk_sector	long	@@@hdsk_sector_
hdsk_track	long	@@@hdsk_track_
hdsk_dma	long	@@@hdsk_dma_

hdsk_dpb	long	@@@hdsk_dpb_

hdsk_dir	long	0
hdsk_params_ptr	long	0		' pointer into hdsk_params in hub RAM
hdsk_params_cnt	long	0		' count of parameters expected

boot0		long	0-0		' logical block address of boot sector
lba0		long	0-0		' logical block address of CP/M
lba		long	0		' logical block translated from track*32+sector
iobuff_lba	long	-1		' logical block that is currently in iobuff
sector_offs	long	0		' offset of current sector in iobuff_ptr
#ifdef TIMED_FLUSH
time_dirty	long	0
#endif TIMED_FLUSH

t1		long	0
t2		long	0
t3		long	0

strptr		long	0
hexval		long	0
hexcnt		long	0

startup
		mov	data, #12				' form feed
		jmp	#output_conout

command
#ifdef TIMED_FLUSH
		mov	t1, time_dirty		WZ
	if_nz	sub	t1, CNT
	if_nz	cmps	t1, #0			WC
  if_nz_and_c	call	#spi_write_block				' yes, write the buffer and clear the dirty flag
#endif TIMED_FLUSH
		rdlong	iocmd, iocmd_ptr
		mov	t1, iocmd
		shr	t1, #24			WZ		' inspect bits 31..24: zero?
	if_z	jmp	#command				' yes, loop waiting for a command.

		mov	port, iocmd				' extract port from bits 23..8
		shr	port, #8				' to bits 15..0
		and	port, #$ff				' mask bits 7..0 of port address

		cmp	t1, #io_cmd_out		WZ
	if_z	jmp	#output
		cmp	t1, #io_cmd_in		WZ
	if_z	jmp	#input
		cmp	t1, #io_cmd_break	WZ
	if_z	jmp	#do_break
command_0
		and	iocmd, low_24bits			' clear command byte in bits 31..24
		wrlong	iocmd, iocmd_ptr
		jmp	#command

'**********************************************************************************************
'
' Read i/o port and data to write to
'
output
		mov	data, iocmd				' extract data
		and	data, #$ff				' mask 8 data bits
		cmp	port, #condata		WZ		' console output?
	if_z	jmp	#output_conout
'		cmp	port, #punstat		WZ		' punch status output - really?
'	if_z	jmp	#output_punstat
		cmp	port, #pundata		WZ		' punch output?
	if_z	jmp	#output_pundata
		cmp	port, #hdsk_port	WZ		' hard disk port?
	if_z	jmp	#ouput_hdsk
		cmp	port, #simh_port 	WZ		' SIMH port?
	if_z	jmp	#ouput_simh
		jmp	#command_0

'**********************************************************************************************
'
' Write data to the console. Writes to the FullDuplexSerial tx_buffer0
'
output_conout
		call	#conout_char
		jmp	#command_0

'**********************************************************************************************
'
' Output a byte in hexval to the console
'
conout_hex_b
		mov	hexcnt, #2
		shl	hexval, #24
		jmp	#conout_hex

'**********************************************************************************************
'
' Output a word in hexval to the console
'
conout_hex_w
		mov	hexcnt, #4
		shl	hexval, #16
		jmp	#conout_hex

'**********************************************************************************************
'
' Output a long in hexval to the console
'
conout_hex_l
		mov	hexcnt, #8
conout_hex
		mov	data, hexval
		shr	data, #28
		cmp	data, #10	WZ, WC
	if_b	add	data, #48
	if_ae	add	data, #55
		call	#conout_char
		shl	hexval, #4
		djnz	hexcnt, #conout_hex
conout_hex_ret
conout_hex_l_ret
conout_hex_w_ret
conout_hex_b_ret
		ret

'**********************************************************************************************
'
' Output a space character to the console
'
conout_space
		mov	data, #32
		call	#conout_char
conout_space_ret
		ret

'**********************************************************************************************
'
' Output a carriage return to the console
'
conout_crlf
		mov	data, #13
		call	#conout_char
		mov	data, #10
		call	#conout_char
conout_crlf_ret
		ret

'**********************************************************************************************
'
' Output a string at strptr to the console
'
conout_string
		rdbyte	data, strptr	WZ
		add	strptr, #1
	if_z	jmp	#conout_string_ret
		call	#conout_char
		jmp	#conout_string
conout_string_ret
		ret

'**********************************************************************************************
'
' Output to the punch status - i.e. punch control - needed?
'
'output_punstat
'		mov	data, #0
'		jmp	#command_0

'**********************************************************************************************
'
' Output to the punch. TODO: serial port 2?
'
output_pundata
		call	#tx1
		jmp	#command_0

'**********************************************************************************************
'
' Handle SIMH compatible command
'
ouput_simh
		tjnz	simh_params_ptr, #output_simh_params

		cmp	data, #simh_reset WZ	' reset what?
	if_z	jmp	#output_simh_reset
		cmp	data, #simh_hasbankedmemory WZ
	if_z	jmp	#output_simh_hasbankedmemory
		cmp	data, #simh_stoptimer WZ
	if_z	jmp	#output_simh_stoptimer
		cmp	data, #simh_setbankselect WZ
	if_z	jmp	#output_simh_setbankselect
		mov	simh_status, #$01	' set error status?!
		jmp	#command_0

output_simh_params
		wrbyte	data, simh_params_ptr
		add	simh_params_ptr, #1
		sub	simh_params_cnt, #1 WZ
	if_nz	jmp	#command_0
		mov	simh_params_ptr, #0	' stop recording parameters
		' TODO: handle simh_command
		cmp	simh_command, #simh_setbankselect WZ
	if_z	jmp	#output_simh_set_bank
		jmp	#command_0

output_simh_reset
		mov	simh_params_ptr, #0	' stop recording or replying parameters
		mov	simh_params_cnt, #0
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command_0

output_simh_setbankselect
		mov	simh_command, data
		mov	simh_params_ptr, #simh_params
		mov	simh_params_cnt, #1	' expect one more output: the bank number
		jmp	#command_0

output_simh_set_bank
		' TODO: adjust RAM address in qz80!?
		jmp	#command_0

output_simh_hasbankedmemory
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command_0

output_simh_stoptimer
		mov	simh_command, data
		mov	simh_status, #0
		jmp	#command_0

'**********************************************************************************************
'
' Handle hard disk command
'
ouput_hdsk
		cmp	hdsk_dir, #dir_out	WZ
	if_z	tjnz	hdsk_params_ptr, #output_hdsk_params

		cmp	data, #hdsk_reset	WZ
	if_z	jmp	#output_hdsk_reset
		cmp	data, #hdsk_read	WZ
	if_z	jmp	#output_hdsk_read
		cmp	data, #hdsk_write	WZ
	if_z	jmp	#output_hdsk_write
		cmp	data, #hdsk_param	WZ
	if_z	jmp	#output_hdsk_drive
		mov	hdsk_status, #$01		' set error status?!
		jmp	#command_0

'**********************************************************************************************
'
' Write port value to RAM at hdsk_params_ptr
'
output_hdsk_params
		wrbyte	data, hdsk_params_ptr
		add	hdsk_params_ptr, #1
		sub	hdsk_params_cnt, #1 WZ
	if_nz	jmp	#command_0
		mov	hdsk_params_ptr, #0		' stop recording parameters

		' handle hdsk_command
		cmp	hdsk_command, #hdsk_param	WZ
	if_z	jmp	#output_hdsk_params_drive
		cmp	hdsk_command, #hdsk_read	WZ
	if_z	jmp	#hdsk_read_dma
		cmp	hdsk_command, #hdsk_write	WZ
	if_z	jmp	#hdsk_write_dma
		jmp	#command_0

'**********************************************************************************************
'
' Reset the hdsk interface
'
output_hdsk_reset
		mov	hdsk_command, data
		mov	hdsk_params_ptr, #0		' stop recording parameters
		mov	hdsk_params_cnt, #0
		mov	hdsk_status, #0
		mov	hdsk_dir, #dir_none		' no direction
		jmp	#command_0

'**********************************************************************************************
'
' Read and write expect 6 additional
' writes to the port
'
output_hdsk_read
output_hdsk_write
		mov	hdsk_command, data
		mov	hdsk_params_ptr, hdsk_drive
		mov	hdsk_params_cnt, #6		' drive, sector, track lo, track hi, DMA lo, DMA hi
		mov	hdsk_status, #1			' non-zero status while expecting parameters?
		mov	hdsk_dir, #dir_out
		jmp	#command_0

'**********************************************************************************************
'
' Params expects one more out, the drive
' number, to be written to the port
'
output_hdsk_drive
		mov	hdsk_command, data
		mov	hdsk_params_ptr, hdsk_drive
		mov	hdsk_params_cnt, #1		' expect drive number
		mov	hdsk_dir, #dir_out
		jmp	#command_0

'**********************************************************************************************
'
' Params returns 19 bytes on the following
' reads from the port
output_hdsk_params_drive
		cmp	data, #8	WZ, WC		' we support drives 0 to 7
	if_ae	jmp	#:error
		mov	hdsk_params_ptr, hdsk_dpb
		mov	hdsk_params_cnt, #19		' 19 bytes to read
		mov	hdsk_status, #0
		mov	hdsk_dir, #dir_in
		jmp	#command_0
:error
		mov	hdsk_status, #$ff
		jmp	#command_0

'**********************************************************************************************
'
' Wait until the XMM driver clears the
' long[pdo_command]
'
pdo_wait_busy
		wrlong	t1, pdo_command			' send the command
:wait		rdlong	t1, pdo_command
		shl	t1, #1		WZ, WC, NR	' look for negative values or zero
  if_nz_and_nc	jmp	#:wait
	if_nz	mov	hdsk_status, t1
	if_z	mov	hdsk_status, #0
pdo_wait_busy_ret
		ret

'**********************************************************************************************
'
' Execute the hdsk read
'
hdsk_read_dma
		call	#spi_read_block
		rdword	t1, hdsk_dma			' get DMA address
		wrlong	t1, pdo_blocknr			' set address of 128 byte sector
		mov	t1, iobuff_ptr			' 512 byte buffer
		add	t1, sector_offs			' + 128 byte sector offset
		or	t1, h128			' 128 bytes
		wrlong	t1, pdo_address			' as XMM transfer length + hub RAM address
		mov	t1, #"W"			' XMM write command
		call	#pdo_wait_busy
		jmp	#command_0

'**********************************************************************************************
'
' Execute the hdsk write
'
hdsk_write_dma
		call	#spi_read_block			' make sure we have the sector in our buffer
		rdword	t1, hdsk_dma			' get DMA address
		wrlong	t1, pdo_blocknr			' set hub address of 128 byte sector
		mov	t1, iobuff_ptr			' 512 byte buffer
		add	t1, sector_offs			' + 128 byte sector offset
		or	t1, h128			' 128 bytes
		wrlong	t1, pdo_address			' as XMM transfer length + RAM address
		mov	t1, #"R"			' XMM read command
		call	#pdo_wait_busy
#ifdef TIMED_FLUSH
		rdlong	time_dirty, #0			' get clockfrq
		add	time_dirty, CNT
#else
		call	#spi_write_block
#endif !TIMED_FLUSH
		jmp	#command_0

'**********************************************************************************************
'
' Read a logical block address from SD
'
spi_read_block
		rdbyte	lba, hdsk_drive			' get drive number
		and	lba, #7				' we support 8 drives, no more
		shl	lba, #2				' * 4
		add	lba, hdsk_lba_drives		' offset of the per drive LBA0
		rdlong	lba0, lba
		rdword	lba, hdsk_track			' lba = track number
		shl	lba, #5				' * 32 sectors per track
		rdbyte	t1, hdsk_sector			' t1 = sector
		add	lba, t1				' + sector number
		mov	sector_offs, lba
		and	sector_offs, #3			' mask sector offset 0 to 3
		shl	sector_offs, #7			' * 128
		shr	lba, #2				' sector number / 4 (read 512 byte sectors)
		add	lba, lba0			' + first sector of this drive
		shl	lba, #8				' make sure bits 31..24 are zero
		shr	lba, #8
		cmp	lba, iobuff_lba		WZ	' == buffered lba?
	if_z	jmp	#spi_read_block_ret		' yes, no need to read again
#ifdef TIMED_FLUSH
		cmp	time_dirty, #0		WZ	' flush a pending write
	if_nz	call	#spi_write_block
#endif TIMED_FLUSH
		mov	iobuff_lba, lba
		wrlong	iobuff_lba, pdo_blocknr
		wrlong	iobuff_ptr, pdo_address
		mov	t1, #"r"			' SD/SPI read command
		call	#pdo_wait_busy
spi_read_block_ret
		ret

'**********************************************************************************************
'
' Write a logical block address to SD
'
spi_write_block
		wrlong	iobuff_lba, pdo_blocknr
		wrlong	iobuff_ptr, pdo_address
		mov	t1, #"w"			' SD/SPI write command
		call	#pdo_wait_busy
#ifdef TIMED_FLUSH
		mov	time_dirty, #0			' buffer is clean
#endif
spi_write_block_ret
		ret


'**********************************************************************************************
'
' Read i/o port to read from
'
input
		andn	iocmd, #$ff			' clear the iocmd data bits 7..0

		cmp	port, #simh_port	WZ
	if_z	jmp	#input_simh
		cmp	port, #hdsk_port	WZ
	if_z	jmp	#input_hdsk
		cmp	port, #constat		WZ
	if_z	jmp	#input_constat
		cmp	port, #condata		WZ
	if_z	jmp	#input_condata
		cmp	port, #punstat		WZ
	if_z	jmp	#input_punstat
		cmp	port, #pundata		WZ
	if_z	jmp	#input_pundata
#ifdef HAVE_KBD
		cmp	port, #keystat		WZ
	if_z	jmp	#input_keystat
		cmp	port, #keydata		WZ
	if_z	jmp	#input_keydata
#endif
		jmp	#command_0

'**********************************************************************************************
'
' Read from the simh port
'
input_simh
		tjnz	simh_params_ptr, #input_simh_params
		or	iocmd, simh_status
		jmp	#command_0

'**********************************************************************************************
'
' Read port value from RAM at simh_params_ptr
'
input_simh_params
		rdbyte	data, simh_params_ptr
		add	simh_params_ptr, #1
		or	iocmd, data
		sub	simh_params_cnt, #1	WZ
	if_z	mov	simh_params_ptr, #0		' stop replying parameters
		jmp	#command_0

'**********************************************************************************************
'
' Read from the hdsk port
'
input_hdsk
		cmp	hdsk_dir, #dir_in	WZ
	if_z	tjnz	hdsk_params_ptr, #input_hdsk_params
		or	iocmd, hdsk_status
		jmp	#command_0

'**********************************************************************************************
'
' Read port value from RAM at hdsk_params_ptr
'
input_hdsk_params
		rdbyte	data, hdsk_params_ptr
		add	hdsk_params_ptr, #1
		or	iocmd, data
		sub	hdsk_params_cnt, #1	WZ
	if_z	mov	hdsk_params_ptr, #0		' stop replying parameters
		jmp	#command_0

'**********************************************************************************************
'
' Read the console status
'
input_constat
		rdlong	t1, rx_head0
		rdlong	t2, rx_tail0
		cmp	t1, t2			WZ
#ifdef HAVE_KBD
		rdlong	t1, kbd_head
		rdlong	t2, kbd_tail
	if_z	cmp	t1, t2			WZ
#endif HAVE_KBD
input_constat_1
	if_z	mov	data, #%00000010	' no data available
	if_nz	mov	data, #%00000011	' data is available
input_constat_exit
		or	iocmd, data
		jmp	#command_0

#ifdef HAVE_KBD
'**********************************************************************************************
'
' Read the keyboard status
'
input_keystat
		rdlong	t1, kbd_head
		rdlong	t2, kbd_tail
		cmp	t1, t2			WZ
		jmp	#input_constat_1
#endif

'**********************************************************************************************
'
' Read the console data
'
' returns 0 if no data is available
'
input_condata
		mov	data, #0		' no data available
		rdlong	t1, rx_head0
		rdlong	t2, rx_tail0
		cmp	t1, t2		WZ
	if_z	jmp	#:kbd
		mov	addr, rx_buffer0
		add	addr, t2
		rdbyte	data, addr
		add	t2, #1
		and	t2, #63
		wrlong	t2, rx_tail0
		jmp	#input_constat_exit
:kbd

#ifdef HAVE_KBD
input_keydata
		rdlong	t1, kbd_head
		rdlong	t2, kbd_tail
		mov	t3, t2
		cmp	t1, t2		WZ
	if_z	jmp	#input_constat_exit

		shl	t2, #1
		add	t2, kbd_keys
		rdword	data, t2		' read just the lower byte
		add	t3, #1
		and	t3, #15
		wrlong	t3, kbd_tail
	
		' process key codes
		test	data, kbd_ctrl	WZ	' is it a control code?
	if_nz	and	data, #$1f		' make it $00..$1f
	if_nz	jmp	#input_constat_exit
		cmp	data, #200	WZ	' backspace?
	if_z	mov	data, #8		' backspace character
	if_z	jmp	#input_constat_exit
		cmp	data, #201	WZ	' is it delete?
	if_z	mov	data, #127		' Delete character
	if_z	jmp	#input_constat_exit
		cmp	data, #203	WZ	' upper code for Esc?
	if_z	mov	data, #27		' Escape character
	if_z	jmp	#input_constat_exit
		cmp	data, #128	WZ, WC	' don't return other codes
	if_ae	mov	data, #0
#endif HAVE_KBD
		jmp	#input_constat_exit

'**********************************************************************************************
'
' Read the paper punch status
'
input_punstat
		mov	data, #%00000011	' always allow to output to punch
		jmp	#input_constat_exit

'**********************************************************************************************
'
' Read the paper punch data
'
' returns 0 if no data is available
'
input_pundata
		mov	data, #0		' no data available
		jmp	#input_constat_exit

'**********************************************************************************************
'
' Break - issued when the Z80 encounters
' a HALT instruction.
'
break
		mov	t1, iocmd			' read the address of the register file
		shr	t1, #8				' bits 23..8 to 15..0
		and	t1, low_16bits			' mask only bits 15..0
		mov	strptr, msg_break_ptr	

		mov	t2, #8
:loop_1		call	#conout_string
		rdword	hexval, t1			' read BC, DE, HL, AF, BC', DE', HL', AF'
		call	#conout_hex_w
		add	t1, #2
		djnz	t2, #:loop_1

		call	#conout_string			' new line and print "R:"
		rdbyte	hexval, t1			' R
		add	t1, #1
		rdbyte	t2, t1				' R2
		and	hexval, #%01111111		' R register's bit #7
		or	hexval, t2			' does never change from the last written value
		call	#conout_hex_b
		add	t1, #1

		mov	t2, #3
		call	#conout_string
		rdbyte	hexval, t1			' IFF1+2/IM
		mov	t2, t1				' save value
		and	hexval, #1			' IFF1 (interrupt flip-flop 1)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #1			' shift bit #1 to #0
		and	hexval, #1			' IFF2 (interrupt flip-flop 2)
		call	#conout_hex_b

		call	#conout_string
		mov	hexval, t2			' IFF1+2/IM
		shr	hexval, #2			' shift bits #3..2 to #1..0
		and	hexval, #3			' IM (interrupt mode)
		call	#conout_hex_b
		add	t1, #1

		call	#conout_string
		rdbyte	hexval, t1			' I (interrupt vector register)
		call	#conout_hex_b
		add	t1, #1

		mov	t2, #4
:loop_2
		call	#conout_string
		rdword	hexval, t1			' IX, IY, SP, PC
		mov	t3, t1
		call	#conout_hex_w
		add	t1, #2
		djnz	t2, #:loop_2

		call	#conout_string

		mov	t2, #4				' 4 bytes
		shl	t2, #16
		or	t2, t3				' length in upper 16 bits, address in lower 16 bits
		wrlong	t2, pdo_address			' as XMM ram address
		mov	t2, #"R"			' XMM read command
		wrlong	t2, pdo_command
:wait		rdlong	t2, pdo_command		WZ, WC
  if_nc_and_nz	jmp	#:wait

		rdbyte	hexval, t1
		call	#conout_hex_b
		call	#conout_crlf
break_ret
		ret

do_break
		call	#break
		jmp	#command_0

conout_char
		call	#tx0
#ifdef HAVE_VGA
		call	#vt
#endif
#ifdef HAVE_TV
		call	#tv
#endif
conout_char_ret
		ret

'**********************************************************************************************
'
' Write to the FullDuplexSerial tx_buffer0
'
tx0
:wait
		rdlong	sio_head, tx_head0
		mov	sio_temp, sio_head
		add	sio_head, #1
		rdlong	sio_tail, tx_tail0
		and	sio_head, #15
		cmp	sio_head, sio_tail	WZ
	if_z	jmp	#:wait					' wait until there's room in the buffer
		mov	addr, tx_buffer0
		add	addr, sio_temp
		wrbyte	data, addr
		wrlong	sio_head, tx_head0			' increment the tx_head0
tx0_ret
		ret
'**********************************************************************************************
'
' Write to the FullDuplexSerial tx_buffer1
'
tx1
:wait
		rdlong	sio_head, tx_head1
		mov	sio_temp, sio_head
		add	sio_head, #1
		rdlong	sio_tail, tx_tail1
		and	sio_head, #15
		cmp	sio_head, sio_tail	WZ
	if_z	jmp	#:wait					' wait until there's room in the buffer
		mov	addr, tx_buffer1
		add	addr, sio_temp
		wrbyte	data, addr
		wrlong	sio_head, tx_head1			' increment the tx_head1
tx1_ret
		ret

#ifdef HAVE_VGA
'**********************************************************************************************
'
' Write to the VT100 on VGA rendezvous long
'
vt
:wait		rdlong	sio_temp, vt_cmdptr	WZ
	if_nz	jmp	#:wait
		mov	sio_temp, #$100
		or	sio_temp, data
		wrlong	sio_temp, vt_cmdptr
vt_ret
		ret
#endif HAVE_VGA

#ifdef HAVE_TV
'**********************************************************************************************
'
' Write to the VT100 on TV rendezvous long
'
tv
:wait		rdlong	sio_temp, vt_cmdptr	WZ
	if_nz	jmp	#:wait
		mov	sio_temp, #$100
		or	sio_temp, data
		wrlong	sio_temp, vt_cmdptr
tv_ret
		ret
#endif HAVE_TV

		long	0[$1f0-$]

		fit	$1f0
#ifdef HAVE_VGA
		long	0[SCREEN_W * SCREEN_H / 4 - $1f0]	' 80x40 = 3200 = $320 longs
#endif HAVE_VGA

#ifdef HAVE_TV
		long	0[SCREEN_W * SCREEN_H / 4 - $1f0]	' 80x25 = 2000 = $1f4 longs
#endif

drive_name	long	@@@drive_name_a, @@@drive_name_b, @@@drive_name_c, @@@drive_name_d
		long	@@@drive_name_e, @@@drive_name_f, @@@drive_name_g, @@@drive_name_h

simh_params_	byte	0,0,0,0,0,0,0,0		' parameters of command

lba_drive_a	long	0			' first sector of drive A:
lba_drive_b	long	0			' first sector of drive B:
lba_drive_c	long	0			' first sector of drive C:
lba_drive_d	long	0			' first sector of drive D:
lba_drive_e	long	0			' first sector of drive E:
lba_drive_f	long	0			' first sector of drive F:
lba_drive_g	long	0			' first sector of drive G:
lba_drive_h	long	0			' first sector of drive H:

hdsk_params_	byte
hdsk_drive_	byte	0
hdsk_sector_	byte	0
hdsk_track_	byte	0, 0
hdsk_dma_	byte	0, 0
hdsk_dummy_	word	0

'Hard Disk Parameter Blocks
'Each DPB is 17 bytes as defined for DPBs in CP/M followed by 2 bytes sector size as used by SIMH
hdsk_dpb_	byte			' 8MByte Default SIMH Altair HDSK params
hdsk_spt_lo	byte	$20		' sectors per track (low byte)
hdsk_spt_hi	byte	$00		' sectors per track (high byte)
hdsk_bsh	byte	$05		' data allocation Block SHift factor
hdsk_blm	byte	$1F		' data allocation block mask
hdsk_exm	byte	$01		' EXtent Mask
hdsk_dsm_lo	byte	$F9		' maximum data block number (low_byte)
hdsk_dsm_hi	byte	$07		' maximum data block number (high_byte)
hdsk_drm_lo	byte	$FF		' total number of directory entries (low byte)
hdsk_drm_hi	byte	$03		' total number of directory entries (high byte)
hdsk_al0	byte	$FF		' determine reserved directory blocks
hdsk_al1	byte	$00		' determine reserved directory blocks
hdsk_cks_lo	byte	$00		' size of directory ChecK vector (low byte)
hdsk_cks_hi	byte	$00		' size of directory ChecK vector (high byte)
hdsk_off_lo	byte	$06		' number of reserved tracks (offset) (low byte)
hdsk_off_hi	byte	$00		' number of reserved tracks (offset) (high byte)
hdsk_psh	byte	$00		' Physical record SHift factor, CP/M 3
hdsk_phm	byte	$00		' Physical record Mask, CP/M 3
hdsk_ss_lo	byte	$80		' Sector Size (low byte)
hdsk_ss_hi	byte	$00		' Sector Size (high byte)
hdsk_filler	byte	$00		' align to word boundary

#ifdef FAT12_SUPPORT
iobuff		long	0[2*512/4]	' I/O buffer (need 2 because of FAT12)
#else
iobuff		long	0[512/4]	' I/O buffer for one 512 byte sector
#endif

boot_name	byte	"BOOT.DSK",0
drive_name_a	byte	"A.DSK",0
drive_name_b	byte	"B.DSK",0
drive_name_c	byte	"C.DSK",0
drive_name_d	byte	"D.DSK",0
drive_name_e	byte	"E.DSK",0
drive_name_f	byte	"F.DSK",0
drive_name_g	byte	"G.DSK",0
drive_name_h	byte	"H.DSK",0

msg_break	byte	"BC:",0
		byte	" DE:",0
		byte	" HL:",0
		byte	" AF:",0
		byte	" BC'",0
		byte	" DE'",0
		byte	" HL'",0
		byte	" AF'",0
		byte	13,10, "R:",0
		byte	" IFF1:",0
		byte	" IFF2:",0
		byte	" IM:",0
		byte	" I:",0
		byte	" IX:",0
		byte	" IY:",0
		byte	" SP:",0
		byte	" PC:",0
		byte	" = ",0

